import sys
import os
import shutil
import re
import argparse
from gitignore_parser import parse_gitignore
from debugprint import Debug

debug = Debug("connoisseur:modifyfilesystem")


leading_slash_re = re.compile(r"^\/")


def delete_path(path):
    """Delete a path regardless of whether a file or directory."""
    debug(f"deleting {path}")
    if os.path.isdir(path):
        shutil.rmtree(path)
    elif os.path.isfile(path):
        os.remove(path)


def copy_path(path, origin, destination):
    extension = leading_slash_re.sub("", path.replace(origin, ""))
    # debug(os.path.split(extension))
    folders = os.path.split(extension)[0]
    dest_folders = os.path.join(destination, folders)
    if not os.path.isdir(dest_folders):
        os.makedirs(dest_folders)
    shutil.copy(path, dest_folders)


def clear_empty_directories(origin):
    for root, dirs, ignored in os.walk(origin):
        for dir in dirs:
            path = os.path.join(root, dir)
            # debug(path, "path")
            # debug(os.listdir(path), "listdir")
            if not os.listdir(path):
                # debug("deleting directory")
                shutil.rmtree(path)



import sys
import os
from debugprint import Debug

from .utils import printif
from . import modifyfilesystem

debug = Debug("connoisseur:tidy")


def _get_files_to_tidy(origin, checker):
    output = []
    for item in map(lambda x: os.path.join(origin, x), os.listdir(origin)):
        if checker.reject(item):
            output.append(item)
        elif os.path.isdir(item):
            output.extend(_get_files_to_tidy(item, checker))
    return output


def tidy(origin, checker, dry_run=True, verbose=False):
    sys.stderr.write("Getting list of files to delete...\n")
    files_to_delete = _get_files_to_tidy(origin, checker)
    sys.stderr.write("Deleting...")
    for path in files_to_delete:
        printif(path, (verbose or dry_run))
        if not dry_run:
            modifyfilesystem.delete_path(path)
    if not dry_run:
        modifyfilesystem.clear_empty_directories(origin)



import os
from gitignore_parser import parse_gitignore
from debugprint import Debug

debug = Debug("connoisseur:checkers")


class CheckerFromRejectFile:
    def __init__(self, file_path, origin_path):
        self.origin = origin_path
        self.checker = parse_gitignore(file_path, origin_path)

    def reject(self, path):
        """Check whether a path should be rejected."""
        return self.checker(path)

    def select(self, path):
        return not self.reject(path)


class CheckerFromSelectFile:
    def __init__(self, file_path, origin_path):
        self.origin = origin_path
        negative_checker = parse_gitignore(file_path, origin_path)
        self.checker = lambda path: not negative_checker(path)

    def reject(self, path):
        """Check whether a path should be rejected."""
        return self.checker(path)

    def select(self, path):
        return not self.reject(path)



import sys
import os
from debugprint import Debug

from .utils import printif, errif
from . import modifyfilesystem

debug = Debug("connoisseur:copy")


def _get_files_to_copy(origin, checker):
    output = []
    for item in map(lambda x: os.path.join(origin, x), os.listdir(origin)):
        if checker.select(item):
            if os.path.isfile(item):
                output.append(item)
            else:
                output.extend(_get_files_to_copy(item, checker))
    return output


def copy(origin, destination, checker, dry_run=True, verbose=False):
    sys.stderr.write("Getting list of files to copy...\n")
    files_to_copy = _get_files_to_copy(origin, checker)
    sys.stderr.write("Copying...\n")
    for path in files_to_copy:
        printif(path, (verbose or dry_run))
        debug("maybe copying...")
        if not dry_run:
            debug("copying...")
            modifyfilesystem.copy_path(path, origin, destination)



import sys
import os
import shutil
import argparse
from debugprint import Debug

from .tidy import tidy
from .copy import copy
from .checkers import CheckerFromRejectFile, CheckerFromSelectFile

debug = Debug("connoisseur:main")


def check_continue(path):
    """Perform a user check whether to continue. Exit if not confirmed."""
    should_continue = input(
        "This will probably delete or overwrite some files at {}. Are you "
        "sure you want to continue? y/N  ".format(path)
    )
    if should_continue not in "yY":
        sys.stderr.write("Not running...connoisseur will now exit.\n")
        sys.exit()


def main():
    parser = argparse.ArgumentParser(
        prog="connoisseur", description="Utility for selective copying and deleting",
    )

    parser.add_argument(
        "action",
        choices=["copy", "tidy"],
        help="Action for connoisseur to perform - tidy up a directory tree or "
        "selectively copy it to a new one.",
    )
    parser.add_argument(
        "spec_file",
        help="Path to the spec file being used. Spec file should be written "
        "in gitignore format.",
    )
    parser.add_argument(
        "origin",
        help="Path to be tidied (in tidy) or used as a source of files to be "
        "copied to the destination (in copy).",
    )
    if len(sys.argv) > 1 and sys.argv[1] == "copy":
        parser.add_argument(
            "destination", help="Destination for files and folders to be copied to."
        )
    parser.add_argument(
        "-s",
        "--spec-type",
        choices=["select", "reject"],
        help="Set spec type - defaults to 'reject' if not specified.",
    )
    parser.add_argument(
        "-d",
        "--dry-run",
        action="store_true",
        help="If specified, carries out a dry run - prints to stdout paths of "
        "all files to be copied (in a copy operation) or deleted (in a tidy "
        "operation) but does not alter the filesystem.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Prints to stdout paths of all files copied (in a copy "
        "operation) or deleted (in a tidy operation).",
    )
    parser.add_argument(
        "-y",
        "--skip-confirmation-check",
        action="store_true",
        help="Skips confirmation check",
    )
    args = parser.parse_args()

    if args.spec_type:
        if args.spec_type == "select":
            checker = CheckerFromSelectFile(args.spec_file, args.origin)
    else:
        checker = CheckerFromRejectFile(args.spec_file, args.origin)

    if args.action == "copy":
        if os.path.isdir(args.destination) and os.listdir(args.destination):
            if not (args.skip_confirmation_check or args.dry_run):
                check_continue(args.destination)
        copy(
            args.origin,
            args.destination,
            checker,
            dry_run=args.dry_run,
            verbose=args.verbose,
        )
    else:
        if not (args.skip_confirmation_check or args.dry_run):
            check_continue(args.origin)
        tidy(args.origin, checker, dry_run=args.dry_run, verbose=args.verbose)



import sys


def errif(string, should_print):
    if should_print:
        sys.stderr.write("{}\n".format(string))


def printif(string, should_print):
    if should_print:
        print(string)


def check_maximum_recursion(fn):
    def recursion_safe_fn(*args):
        try:
            return fn(*args)
        except RecursionError:
            sys.stderr.write(
                "Error: maximum recursion depth exceeded. You are probably "
                "attempting to copy or tidy a very deeply nested directory "
                "tree (hundreds or thousands of directories deep). "
                "Unfortunately, the current implementation of Connoisseur "
                "isn't able to handle directories nested to that level as "
                "Python doesn't support tail-call recursion.\nConnoisseur "
                "will now exit.\n",
            )
